# pylint: disable=global-statement
import json
import os
import signal
import sys
from textwrap import dedent
from time import sleep
from typing import List

import pycurl

from pcs.cli.reports.processor import print_report
from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
    TaskIdentDto,
    TaskResultDto,
)
from pcs.common.async_tasks.types import (
    TaskFinishType,
    TaskState,
)
from pcs.common.interface.dto import (
    from_dict,
    to_dict,
)
from pcs.common.node_communicator import (
    HostNotFound,
    NodeTargetFactory,
)
from pcs.common.reports import ReportItemDto
from pcs.utils import read_known_hosts_file

task_ident = ""
report_list: List[ReportItemDto] = []
kill_requested = False


def get_token_for_localhost() -> str:
    try:
        return (
            NodeTargetFactory(read_known_hosts_file())
            .get_target("localhost")
            .token
        )
    except HostNotFound as e:
        error(f"No token found for '{e.name}'")
        raise SystemExit(1) from e


def get_signal_handler(token: str):
    def signal_handler(sig, frame):
        # pylint: disable=unused-argument
        if sig == signal.SIGINT:
            global kill_requested
            if not task_ident:
                error("no task to kill")
                raise SystemExit(1) from None
            if kill_requested:
                raise SystemExit(1)

            kill_requested = True
            # Kill task request
            task_ident_dto = TaskIdentDto(task_ident)
            make_api_request_post(
                "task/kill", json.dumps(to_dict(task_ident_dto)), token
            )
            print("Task kill request sent...")
            task_result_dto = fetch_task_result(task_ident_dto, token)
            print_command_return_value(task_result_dto)
            print_task_details(task_result_dto)
            raise SystemExit(0)

    return signal_handler


# REQUEST FUNCTIONS
def _handle_api_error(curl: pycurl.Curl, response: str) -> None:
    if curl.getinfo(pycurl.RESPONSE_CODE) != 200:
        try:
            error_response = json.loads(response)
            error(error_response.error_message)
        except AttributeError:
            error(response)
        curl.close()
        raise SystemExit(1)
    curl.close()


def _make_api_request(endpoint) -> pycurl.Curl:
    curl = pycurl.Curl()
    curl.setopt(pycurl.URL, f"https://localhost:2224/api/v2/{endpoint}")
    curl.setopt(pycurl.SSL_VERIFYPEER, 0)
    curl.setopt(pycurl.SSL_VERIFYHOST, 0)
    return curl


def make_api_request_get(endpoint: str, params: str, auth_token: str) -> str:
    curl = _make_api_request(endpoint + params)
    curl.setopt(pycurl.COOKIE, f"token={auth_token};".encode("utf-8"))
    response = curl.perform_rs()
    _handle_api_error(curl, response)
    return response


def make_api_request_post(
    endpoint: str,
    json_body: str,
    auth_token: str,
) -> str:
    curl = _make_api_request(endpoint)
    curl.setopt(pycurl.HTTPHEADER, ["Content-Type: application/json"])
    curl.setopt(pycurl.POSTFIELDS, json_body)
    curl.setopt(pycurl.COOKIE, f"token={auth_token};".encode("utf-8"))
    response = curl.perform_rs()
    _handle_api_error(curl, response)
    return response


# PRETTY PRINT
def print_command_return_value(task_result_dto: TaskResultDto) -> None:
    print(task_result_dto.result)


def print_task_details(task_result_dto: TaskResultDto) -> None:
    print(
        dedent(
            f"""
                --------------------------
                Task ident: {task_result_dto.task_ident}
                Task finish type: {task_result_dto.task_finish_type}
                Task kill reason: {task_result_dto.kill_reason}
                """
        )
    )


def error(text: str) -> None:
    print(f"Error: {text}")


# COMMAND CALL
def fetch_task_result(
    task_ident_dto: TaskIdentDto, auth_token: str, sleep_interval: float = 0.3
) -> TaskResultDto:
    task_state = TaskState.CREATED
    # Using global report list to recall reports in signal handler
    global report_list
    while task_state != TaskState.FINISHED:
        response = make_api_request_get(
            "task/result",
            f"?task_ident={task_ident_dto.task_ident}",
            auth_token,
        )
        task_result_dto = from_dict(TaskResultDto, json.loads(response))
        task_state = task_result_dto.state

        # Only print new reports - picks only reports not in report_list
        for report_item_dto in task_result_dto.reports[
            len(report_list) : len(task_result_dto.reports) - len(report_list)
        ]:
            print_report(report_item_dto)
        report_list = task_result_dto.reports
        # Wait for updates and continue until the task is finished
        sleep(sleep_interval)  # 300ms
    global task_ident
    task_ident = ""
    return task_result_dto


def perform_command(command_dto: CommandDto, auth_token: str) -> TaskResultDto:
    response = make_api_request_post(
        "task/create", json.dumps(to_dict(command_dto)), auth_token
    )
    task_ident_dto = from_dict(TaskIdentDto, json.loads(response))
    global task_ident
    task_ident = task_ident_dto.task_ident

    task_result_dto = fetch_task_result(task_ident_dto, auth_token)

    return task_result_dto


def run_command_synchronously(
    command_dto: CommandDto, auth_token: str
) -> TaskResultDto:
    print("Running command synchronously")
    response = make_api_request_post(
        "task/run", json.dumps(to_dict(command_dto)), auth_token
    )
    task_result_dto = from_dict(TaskResultDto, json.loads(response))
    for report_item_dto in task_result_dto.reports:
        print_report(report_item_dto)
    return task_result_dto


def main():
    entrypoint = sys.argv.pop(0)
    if len(sys.argv) not in (1, 2, 3):
        error(f"Usage: {entrypoint} [--sync] <command> [<payload>]")
        raise SystemExit(1)
    run_fn = perform_command
    if sys.argv[0] == "--sync":
        sys.argv.pop(0)
        run_fn = run_command_synchronously
    if len(sys.argv) == 2:
        payload = sys.argv[1]
    else:
        payload = sys.stdin.read()
    token = os.environ.get("PCS_TOKEN", None)
    if token is None:
        token = get_token_for_localhost()
    signal.signal(signal.SIGINT, get_signal_handler(token))
    result = run_fn(
        CommandDto(
            sys.argv[0],
            json.loads(payload),
            options=CommandOptionsDto(
                request_timeout=None,
                # effective_username="custom_user",
                # effective_groups=["custom_group1", "custom_group2"],
            ),
        ),
        token,
    )
    print_command_return_value(result)
    print_task_details(result)
    if result.task_finish_type != TaskFinishType.SUCCESS:
        raise SystemExit(1)



from collections import defaultdict
from dataclasses import dataclass
from typing import (
    AbstractSet,
    List,
    Mapping,
    NewType,
    Optional,
)

from pcs.common.resource_agent.dto import (
    ResourceAgentActionDto,
    ResourceAgentMetadataDto,
    ResourceAgentNameDto,
    ResourceAgentParameterDto,
)

FakeAgentName = NewType("FakeAgentName", str)
OcfVersion = NewType("OcfVersion", str)
_FAKE_AGENT_STANDARD = "__pcmk_internal"


@dataclass(frozen=True)
class ResourceAgentName:
    standard: str
    provider: Optional[str]
    type: str

    @property
    def full_name(self) -> str:
        return ":".join(filter(None, [self.standard, self.provider, self.type]))

    @property
    def is_pcmk_fake_agent(self) -> bool:
        return self.standard == _FAKE_AGENT_STANDARD

    @property
    def is_stonith(self) -> bool:
        return self.standard == "stonith"

    def to_dto(self) -> ResourceAgentNameDto:
        return ResourceAgentNameDto(
            standard=self.standard,
            provider=self.provider,
            type=self.type,
        )

    @classmethod
    def from_dto(cls, dto: ResourceAgentNameDto) -> "ResourceAgentName":
        return cls(
            dto.standard,
            dto.provider,
            dto.type,
        )


@dataclass(frozen=True)
class ResourceAgentActionOcf1_0:  # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes

    # (start, stop, promote...), mandatory by both OCF 1.0 and 1.1
    name: str
    # mandatory by both OCF 1.0 and 1.1, sometimes not defined by agents
    timeout: Optional[str]
    # optional by both OCF 1.0 and 1.1
    interval: Optional[str]
    # optional by OCF 1.1
    # not allowed by OCF 1.0, defined in OCF 1.0 agents anyway
    role: Optional[str]
    # OCF name: 'start-delay', optional by both OCF 1.0 and 1.1
    start_delay: Optional[str]
    # optional by both OCF 1.0 and 1.1
    depth: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    automatic: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    on_target: Optional[str]


@dataclass(frozen=True)
class ResourceAgentActionOcf1_1:  # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes

    # (start, stop, promote...), mandatory by both OCF 1.0 and 1.1
    name: str
    # mandatory by both OCF 1.0 and 1.1, sometimes not defined by agents
    timeout: Optional[str]
    # optional by both OCF 1.0 and 1.1
    interval: Optional[str]
    # optional by OCF 1.1
    role: Optional[str]
    # OCF name: 'start-delay', optional by both OCF 1.0 and 1.1
    start_delay: Optional[str]
    # optional by both OCF 1.0 and 1.1
    depth: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway, most probably
    # will be used in OCF 1.1 agents as well as it holds important information
    automatic: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway, most probably
    # will be used in OCF 1.1 agents as well as it holds important information
    on_target: Optional[str]


@dataclass(frozen=True)
class ResourceAgentParameterOcf1_0:  # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes

    # name of the parameter
    name: str
    # short description
    shortdesc: Optional[str]
    # long description
    longdesc: Optional[str]
    # data type of the parameter
    type: str
    # default value of the parameter
    default: Optional[str]
    # allowed values, only defined if type == 'select'
    enum_values: Optional[List[str]]
    # is this a required parameter?
    required: Optional[str]
    # is this parameter deprecated?
    deprecated: Optional[str]
    # name of a deprecated parameter obsoleted by this one
    obsoletes: Optional[str]
    # should the parameter's value be unique across same agent resources?
    unique: Optional[str]


@dataclass(frozen=True)
class ResourceAgentParameterOcf1_1:  # pylint: disable=invalid-name
    # pylint: disable=too-many-instance-attributes

    # name of the parameter
    name: str
    # short description
    shortdesc: Optional[str]
    # long description
    longdesc: Optional[str]
    # data type of the parameter
    type: str
    # default value of the parameter
    default: Optional[str]
    # allowed values, only defined if type == 'select'
    enum_values: Optional[List[str]]
    # is this a required parameter?
    required: Optional[str]
    # is this parameter deprecated?
    deprecated: bool
    # list of parameters deprecating this one
    deprecated_by: List[str]
    # text describing / explaining the deprecation
    deprecated_desc: Optional[str]
    # should the parameter's value be unique across same agent resources?
    # OCF 1.1 defines "unique" as well, but it is deprecated and we ignore it
    unique_group: Optional[str]
    # changing this parameter's value triggers a reload instead of a restart
    reloadable: Optional[str]


@dataclass(frozen=True)
class ResourceAgentMetadataOcf1_0:  # pylint: disable=invalid-name
    name: ResourceAgentName
    shortdesc: Optional[str]
    longdesc: Optional[str]
    parameters: List[ResourceAgentParameterOcf1_0]
    actions: List[ResourceAgentActionOcf1_0]


@dataclass(frozen=True)
class ResourceAgentMetadataOcf1_1:  # pylint: disable=invalid-name
    name: ResourceAgentName
    shortdesc: Optional[str]
    longdesc: Optional[str]
    parameters: List[ResourceAgentParameterOcf1_1]
    actions: List[ResourceAgentActionOcf1_1]


@dataclass(frozen=True)
class ResourceAgentAction:
    # pylint: disable=too-many-instance-attributes

    # (start, stop, promote...), mandatory by both OCF 1.0 and 1.1
    name: str
    # mandatory by both OCF 1.0 and 1.1, sometimes not defined by agents
    timeout: Optional[str]
    # optional by both OCF 1.0 and 1.1
    interval: Optional[str]
    # optional by OCF 1.1
    # not allowed by OCF 1.0, defined in OCF 1.0 agents anyway
    role: Optional[str]
    # OCF name: 'start-delay', optional by both OCF 1.0 and 1.1
    start_delay: Optional[str]
    # optional by both OCF 1.0 and 1.1
    depth: Optional[str]
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    automatic: bool
    # not allowed by any OCF, defined in OCF 1.0 agents anyway
    on_target: bool

    def to_dto(self) -> ResourceAgentActionDto:
        return ResourceAgentActionDto(
            self.name,
            self.timeout,
            self.interval,
            self.role,
            self.start_delay,
            self.depth,
            self.automatic,
            self.on_target,
        )


@dataclass(frozen=True)
class ResourceAgentParameter:
    # pylint: disable=too-many-instance-attributes

    # name of the parameter
    name: str
    # short description
    shortdesc: Optional[str]
    # long description
    longdesc: Optional[str]
    # data type of the parameter
    type: str
    # default value of the parameter
    default: Optional[str]
    # allowed values, only defined if type == 'select'
    enum_values: Optional[List[str]]
    # True if it is a required parameter, False otherwise
    required: bool
    # True if the parameter is meant for advanced users
    advanced: bool
    # True if the parameter is deprecated, False otherwise
    deprecated: bool
    # list of parameters deprecating this one
    deprecated_by: List[str]
    # text describing / explaining the deprecation
    deprecated_desc: Optional[str]
    # should the parameter's value be unique across same agent resources?
    unique_group: Optional[str]
    # changing this parameter's value triggers a reload instead of a restart
    reloadable: bool

    def to_dto(self) -> ResourceAgentParameterDto:
        return ResourceAgentParameterDto(
            self.name,
            self.shortdesc,
            self.longdesc,
            self.type,
            self.default,
            self.enum_values,
            self.required,
            self.advanced,
            self.deprecated,
            self.deprecated_by,
            self.deprecated_desc,
            self.unique_group,
            self.reloadable,
        )


@dataclass(frozen=True)
class ResourceAgentMetadata:
    name: ResourceAgentName
    agent_exists: bool
    ocf_version: OcfVersion
    shortdesc: Optional[str]
    longdesc: Optional[str]
    parameters: List[ResourceAgentParameter]
    actions: List[ResourceAgentAction]

    @property
    def provides_unfencing(self) -> bool:
        if not self.name.is_stonith:
            return False
        for action in self.actions:
            if action.name == "on" and action.on_target and action.automatic:
                return True
        return False

    @property
    def provides_self_validation(self) -> bool:
        return any(action.name == "validate-all" for action in self.actions)

    @property
    def provides_promotability(self) -> bool:
        return set(action.name for action in self.actions) >= {
            "promote",
            "demote",
        }

    @property
    def unique_parameter_groups(self) -> Mapping[str, AbstractSet[str]]:
        result = defaultdict(set)
        for param in self.parameters:
            if param.unique_group:
                result[param.unique_group].add(param.name)
        return dict(result)

    def to_dto(self) -> ResourceAgentMetadataDto:
        return ResourceAgentMetadataDto(
            self.name.to_dto(),
            self.shortdesc,
            self.longdesc,
            [parameter.to_dto() for parameter in self.parameters],
            [action.to_dto() for action in self.actions],
        )


@dataclass(frozen=True, order=True)
class StandardProviderTuple:
    standard: str
    provider: Optional[str] = None

    @property
    def is_stonith(self) -> bool:
        return self.standard == "stonith"



import json
from typing import Optional

from pcs.common import file_type_codes as code
from pcs.common import reports
from pcs.common.interface.dto import SerializableType
from pcs.lib.interface.config import (
    ExporterInterface,
    ParserErrorException,
    ParserInterface,
)


class JsonParserException(ParserErrorException):
    def __init__(self, json_exception: json.JSONDecodeError):
        super().__init__()
        self.json_exception = json_exception


class JsonParser(ParserInterface):
    """
    Adapts standard json parser to our interfaces
    """

    @staticmethod
    def parse(raw_file_data: bytes) -> SerializableType:
        try:
            # json.loads handles bytes, it expects utf-8, 16 or 32 encoding
            return json.loads(raw_file_data)
        except json.JSONDecodeError as e:
            raise JsonParserException(e) from e

    @staticmethod
    def exception_to_report_list(
        exception: ParserErrorException,
        file_type_code: code.FileTypeCode,
        file_path: Optional[str],
        force_code: Optional[reports.types.ForceCode],
        is_forced_or_warning: bool,
    ) -> reports.ReportItemList:
        if isinstance(exception, JsonParserException):
            if isinstance(exception.json_exception, json.JSONDecodeError):
                return [
                    reports.ReportItem(
                        severity=reports.item.get_severity(
                            force_code,
                            is_forced_or_warning,
                        ),
                        message=reports.messages.ParseErrorJsonFile(
                            file_type_code,
                            exception.json_exception.lineno,
                            exception.json_exception.colno,
                            exception.json_exception.pos,
                            exception.json_exception.msg,
                            str(exception.json_exception),
                            file_path=file_path,
                        ),
                    )
                ]
        raise exception


class JsonExporter(ExporterInterface):
    """
    Adapts standard json exporter to our interfaces
    """

    @staticmethod
    def export(config_structure: SerializableType) -> bytes:
        return json.dumps(
            config_structure,
            indent=4,
            sort_keys=True,
        ).encode("utf-8")



from unittest import TestCase

from lxml import etree

from pcs.lib.cib.resource import clone
from pcs.lib.cib.tools import IdProvider

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)


class AppendNewCommon(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <cib>
                <resources>
                    <primitive id="R"/>
                </resources>
            </cib>
        """
        )
        self.resources = self.cib.find(".//resources")
        self.primitive = self.cib.find(".//primitive")

    def assert_clone_effect(self, options, xml, clone_id=None):
        clone.append_new(
            self.resources,
            IdProvider(self.resources),
            self.primitive,
            options,
            clone_id=clone_id,
        )
        assert_xml_equal(etree.tostring(self.cib).decode(), xml)

    def test_add_without_options(self):
        self.assert_clone_effect(
            {},
            """
            <cib>
                <resources>
                    <clone id="R-clone">
                        <primitive id="R"/>
                    </clone>
                </resources>
            </cib>
        """,
        )

    def test_add_with_options(self):
        self.assert_clone_effect(
            {"a": "b"},
            """
            <cib>
                <resources>
                    <clone id="R-clone">
                        <primitive id="R"/>
                        <meta_attributes id="R-clone-meta_attributes">
                            <nvpair id="R-clone-meta_attributes-a"
                                name="a" value="b"
                            />
                        </meta_attributes>
                    </clone>
                </resources>
            </cib>
        """,
        )

    def test_add_without_options_and_with_custom_id(self):
        self.assert_clone_effect(
            {},
            """
            <cib>
                <resources>
                    <clone id="MyCustomCloneId">
                        <primitive id="R"/>
                    </clone>
                </resources>
            </cib>
            """,
            clone_id="MyCustomCloneId",
        )

    def test_add_with_options_and_with_custom_id(self):
        self.assert_clone_effect(
            {"a": "b"},
            """
            <cib>
                <resources>
                    <clone id="MyCustomCloneId">
                        <primitive id="R"/>
                        <meta_attributes id="MyCustomCloneId-meta_attributes">
                            <nvpair id="MyCustomCloneId-meta_attributes-a"
                                name="a" value="b"
                            />
                        </meta_attributes>
                    </clone>
                </resources>
            </cib>
            """,
            clone_id="MyCustomCloneId",
        )


class IsAnyClone(TestCase):
    def test_is_clone(self):
        self.assertTrue(clone.is_clone(etree.fromstring("<clone/>")))
        self.assertFalse(clone.is_clone(etree.fromstring("<master/>")))
        self.assertFalse(clone.is_clone(etree.fromstring("<group/>")))

    def test_is_master(self):
        self.assertTrue(clone.is_master(etree.fromstring("<master/>")))
        self.assertFalse(clone.is_master(etree.fromstring("<clone/>")))
        self.assertFalse(clone.is_master(etree.fromstring("<group/>")))

    def test_is_any_clone(self):
        self.assertTrue(clone.is_any_clone(etree.fromstring("<clone/>")))
        self.assertTrue(clone.is_any_clone(etree.fromstring("<master/>")))
        self.assertFalse(clone.is_any_clone(etree.fromstring("<group/>")))


class IsPromotableClone(TestCase):
    def my_assert(self, result, xml):
        self.assertEqual(
            result, clone.is_promotable_clone(etree.fromstring(xml))
        )

    def test_master(self):
        self.my_assert(False, "<master />")

    def test_master_with_promotable(self):
        self.my_assert(
            False,
            """
                <master>
                    <meta_attributes>
                        <nvpair name="promotable" value="true" />
                    </meta_attributes>
                </master>
            """,
        )

    def test_clone(self):
        self.my_assert(False, "<clone />")

    def test_clone_with_promotable(self):
        self.my_assert(
            True,
            """
                <clone>
                    <meta_attributes>
                        <nvpair name="promotable" value="true" />
                    </meta_attributes>
                </clone>
            """,
        )

    def test_clone_with_promotable_false(self):
        self.my_assert(
            False,
            """
                <clone>
                    <meta_attributes>
                        <nvpair name="promotable" value="false" />
                    </meta_attributes>
                </clone>
            """,
        )

    def test_clone_with_promotable_in_resource(self):
        self.my_assert(
            False,
            """
                <clone>
                    <primitive>
                        <meta_attributes>
                            <nvpair name="promotable" value="true" />
                        </meta_attributes>
                    </primitive>
                </clone>
            """,
        )


class GetParentAnyClone(TestCase):
    cib = etree.fromstring(
        """
        <cib>
            <resources>
                <primitive id="A" />
                <clone id="B-clone">
                    <primitive id="B" />
                </clone>
                <master id="C-master">
                    <primitive id="C" />
                </master>
                <group id="D">
                    <primitive id="D1" />
                    <primitive id="D2" />
                </group>
                <clone id="E-clone">
                    <group id="E">
                        <primitive id="E1" />
                        <primitive id="E2" />
                    </group>
                </clone>
                <master id="F-master">
                    <group id="F">
                        <primitive id="F1" />
                        <primitive id="F2" />
                    </group>
                </master>
                <bundle id="G-bundle" />
                <bundle id="H-bundle">
                    <primitive id="H" />
                </bundle>
            </resources>
        </cib>
    """
    )

    def my_assert(self, id_in, id_out):
        element_in = self.cib.find(f'.//*[@id="{id_in}"]')
        actual_out = clone.get_parent_any_clone(element_in)
        if id_out is None:
            self.assertIsNone(actual_out)
            return
        if actual_out is None:
            self.assertIsNone(id_out)
            return
        self.assertEqual(id_out, actual_out.get("id"))

    def test_primitive(self):
        self.my_assert("A", None)

    def test_clone(self):
        self.my_assert("B-clone", None)
        self.my_assert("E-clone", None)

    def test_primitive_in_clone(self):
        self.my_assert("B", "B-clone")

    def test_master(self):
        self.my_assert("C-master", None)
        self.my_assert("F-master", None)

    def test_primitive_in_master(self):
        self.my_assert("C", "C-master")

    def test_group(self):
        self.my_assert("D", None)

    def test_primitive_in_group(self):
        self.my_assert("D1", None)

    def test_group_in_clone(self):
        self.my_assert("E", "E-clone")

    def test_primitive_in_group_in_clone(self):
        self.my_assert("E1", "E-clone")

    def test_group_in_master(self):
        self.my_assert("F", "F-master")

    def test_primitive_in_group_in_master(self):
        self.my_assert("F1", "F-master")

    def test_bundle(self):
        self.my_assert("G-bundle", None)
        self.my_assert("H-bundle", None)

    def test_primitive_in_bundle(self):
        self.my_assert("H", None)


class GetInnerResource(TestCase):
    def assert_inner_resource(self, resource_id, xml):
        self.assertEqual(
            resource_id,
            clone.get_inner_resource(etree.fromstring(xml)).get("id", ""),
        )

    def test_primitive(self):
        self.assert_inner_resource(
            "A",
            """
                <clone id="A-clone">
                    <meta_attributes />
                    <primitive id="A" />
                    <meta_attributes />
                </clone>
            """,
        )

    def test_group(self):
        self.assert_inner_resource(
            "A",
            """
                <clone id="A-clone">
                    <meta_attributes />
                    <group id="A" />
                    <meta_attributes />
                </clone>
            """,
        )


class ValidateCloneId(TestCase):
    def setUp(self):
        self.cib = etree.fromstring(
            """
            <cib>
                <resources>
                    <clone id="CloneId">
                        <meta_attributes id="CloneId-meta_attributes"/>
                    </clone>
                </resources>
            </cib>
            """
        )
        self.resources = self.cib.find(".//resources")
        self.id_provider = IdProvider(self.resources)

    def assert_validate_clone_id(self, clone_id, expected_report_item_list):
        assert_report_item_list_equal(
            clone.validate_clone_id(clone_id, self.id_provider),
            expected_report_item_list,
        )

    def test_valid_id(self):
        self.assert_validate_clone_id("UniqueCloneId", [])

    def test_invalid_id_character(self):
        self.assert_validate_clone_id(
            "0CloneId",
            [fixture.report_invalid_id("0CloneId", "0")],
        )

    def test_clone_id_exist(self):
        self.assert_validate_clone_id(
            "CloneId-meta_attributes",
            [fixture.report_id_already_exist("CloneId-meta_attributes")],
        )



import logging
from unittest import mock

from pcs.daemon.app import auth
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser

from pcs_test.tier0.daemon.app import fixtures_app

# Don't write errors to test output.
logging.getLogger("tornado.access").setLevel(logging.CRITICAL)


class AppTest(fixtures_app.AppTest):
    def setUp(self):
        self.auth_provider = AuthProvider(logging.getLogger("test logger"))
        super().setUp()

    def _mock_auth_provider_method(self, method_name, return_value=None):
        method_patcher = mock.patch.object(AuthProvider, method_name)
        self.addCleanup(method_patcher.stop)
        method_mock = method_patcher.start()
        if return_value:
            method_mock.return_value = return_value
        return method_mock

    def get_routes(self):
        return auth.get_routes(
            self.auth_provider,
        )


class Auth(AppTest):
    # pylint: disable=too-many-ancestors
    def setUp(self):
        super().setUp()
        self.auth_provider_mock = self._mock_auth_provider_method(
            "auth_by_username_password"
        )
        self.token = "new token"
        self._mock_auth_provider_method("create_token", self.token)

    def make_auth_request(self):
        return self.post(
            "/remote/auth",
            body={
                "username": fixtures_app.USER,
                "password": fixtures_app.PASSWORD,
            },
        )

    def test_refuse_unknown_user(self):
        self.auth_provider_mock.return_value = None
        self.assertEqual(b"", self.make_auth_request().body)
        self.auth_provider_mock.assert_called_once_with(
            fixtures_app.USER, fixtures_app.PASSWORD
        )

    def test_success(self):
        self.auth_provider_mock.return_value = AuthUser(
            fixtures_app.USER, fixtures_app.GROUPS
        )
        self.assertEqual(
            self.token.encode("utf-8"), self.make_auth_request().body
        )
        self.auth_provider_mock.assert_called_once_with(
            fixtures_app.USER, fixtures_app.PASSWORD
        )



import json
from dataclasses import asdict
from unittest import TestCase

from pcs.common.permissions.types import (
    PermissionAccessType,
    PermissionTargetType,
)
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.permissions.config.parser import (
    ParserError,
    ParserV2,
)
from pcs.lib.permissions.config.types import (
    ClusterEntry,
    ClusterPermissions,
    ConfigV2,
    PermissionEntry,
)


class ParserV2Test(TestCase):
    @staticmethod
    def _run_parse(data):
        return ParserV2.parse(json.dumps(data).encode())

    def test_not_json(self):
        with self.assertRaises(ParserErrorException):
            ParserV2.parse("not json".encode())

    def test_root_element_not_an_object(self):
        with self.assertRaises(ParserError):
            self._run_parse([dict(key="value"), "string", 1])

    def test_format_version_missing(self):
        with self.assertRaises(ParserError):
            self._run_parse(
                asdict(
                    ConfigV2(
                        data_version=1,
                        clusters=[],
                        permissions=ClusterPermissions(local_cluster=[]),
                    )
                )
            )

    def test_format_version_unsupported(self):
        with self.assertRaises(ParserError):
            self._run_parse(
                dict(
                    **asdict(
                        ConfigV2(
                            data_version=1,
                            clusters=[],
                            permissions=ClusterPermissions(local_cluster=[]),
                        )
                    ),
                    format_version=3,
                )
            )

    def test_invalid_data(self):
        with self.assertRaises(ParserError):
            self._run_parse(
                dict(
                    **asdict(
                        ConfigV2(
                            data_version=1,
                            clusters=[
                                ClusterEntry(
                                    name="testcluster",
                                    nodes=["node1", "node2", "node0"],
                                ),
                                dict(name="invalid data"),
                            ],
                            permissions=ClusterPermissions(local_cluster=[]),
                        )
                    ),
                    format_version=2,
                )
            )

    def test_valid_data(self):
        config = ConfigV2(
            data_version=1,
            clusters=[
                ClusterEntry(
                    name="testcluster",
                    nodes=["node1", "node2", "node0"],
                ),
                ClusterEntry(
                    name="cluster name",
                    nodes=["n1", "n0"],
                ),
            ],
            permissions=ClusterPermissions(
                local_cluster=[
                    PermissionEntry(
                        name="user1",
                        type=PermissionTargetType.USER,
                        allow=[PermissionAccessType.READ],
                    ),
                    PermissionEntry(
                        name="group1",
                        type=PermissionTargetType.GROUP,
                        allow=[
                            PermissionAccessType.GRANT,
                            PermissionAccessType.WRITE,
                        ],
                    ),
                ]
            ),
        )
        self.assertEqual(
            self._run_parse(dict(**asdict(config), format_version=2)), config
        )



from .types import CommunicationResultStatus as Status

COM_STATUS_SUCCESS = Status("success")
COM_STATUS_INPUT_ERROR = Status("input_error")
COM_STATUS_UNKNOWN_CMD = Status("unknown_cmd")
COM_STATUS_ERROR = Status("error")
COM_STATUS_EXCEPTION = Status("exception")
COM_STATUS_NOT_AUTHORIZED = Status("not_authorized")



from .expression_part import (
    NODE_ATTR_TYPE_INTEGER,
    BoolExpr,
    NodeAttrExpr,
    OpExpr,
    RscExpr,
    RuleExprPart,
)


def has_node_attr_expr_with_type_integer(rule_expr: RuleExprPart) -> bool:
    if (
        isinstance(rule_expr, NodeAttrExpr)
        and rule_expr.attr_type == NODE_ATTR_TYPE_INTEGER
    ):
        return True
    if isinstance(rule_expr, BoolExpr):
        return any(
            has_node_attr_expr_with_type_integer(child)
            for child in rule_expr.children
        )
    return False


def has_rsc_or_op_expression(rule_expr: RuleExprPart) -> bool:
    if isinstance(rule_expr, OpExpr):
        return True
    if isinstance(rule_expr, RscExpr):
        return True
    if isinstance(rule_expr, BoolExpr):
        return any(
            has_rsc_or_op_expression(child) for child in rule_expr.children
        )
    return False



import logging
from unittest import mock

from pcs.daemon import ruby_pcsd
from pcs.daemon.app import sinatra_ui
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser

from pcs_test.tier0.daemon.app import fixtures_app

# Don't write errors to test output.
logging.getLogger("tornado.access").setLevel(logging.CRITICAL)


class AppTest(fixtures_app.AppUiTestMixin):
    def setUp(self):
        self.wrapper = fixtures_app.RubyPcsdWrapper(ruby_pcsd.SINATRA)
        super().setUp()

    def get_routes(self):
        return sinatra_ui.get_routes(
            self.session_storage,
            AuthProvider(logging.getLogger("test logger")),
            self.wrapper,
        )


@mock.patch.object(
    AuthProvider,
    "login_user",
    lambda _self, username: AuthUser(username=username, groups=[username]),
)
class SinatraAjaxProtected(AppTest):
    # pylint: disable=too-many-ancestors
    def test_deal_without_authentication(self):
        self.assert_unauth_ajax(self.get("/some-ajax", is_ajax=True))

    def test_not_ajax(self):
        session1 = self.create_login_session()
        self.assert_unauth_ajax(self.get("/some-ajax", sid=session1.sid))

    def test_take_result_from_ruby(self):
        session1 = self.create_login_session()
        response = self.get("/some-ajax", sid=session1.sid, is_ajax=True)
        self.assert_wrappers_response(response)
        self.assert_session_in_response(response, session1.sid)



from logging import Logger
from typing import (
    Any,
    Callable,
    Mapping,
    Optional,
    Union,
    cast,
)

from lxml.etree import _Element

from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.host import PcsKnownHost
from pcs.common.node_communicator import (
    Communicator,
    NodeCommunicatorFactory,
)
from pcs.common.reports import ReportProcessor
from pcs.common.reports.item import ReportItem
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.common.tools import Version
from pcs.common.types import StringIterable
from pcs.lib.booth.env import BoothEnv
from pcs.lib.communication import qdevice
from pcs.lib.communication.corosync import (
    CheckCorosyncOffline,
    DistributeCorosyncConf,
    ReloadCorosyncConf,
)
from pcs.lib.communication.tools import (
    run,
    run_and_raise,
)
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfigFacade
from pcs.lib.corosync.config_parser import (
    verify_section as verify_corosync_section,
)
from pcs.lib.corosync.live import get_local_corosync_conf
from pcs.lib.dr.env import DrEnv
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.file.instance import FileInstance
from pcs.lib.interface.config import ParserErrorException
from pcs.lib.node import get_existing_nodes_names
from pcs.lib.node_communication import (
    LibCommunicatorLogger,
    NodeTargetLibFactory,
)
from pcs.lib.pacemaker.live import (
    diff_cibs_xml,
    ensure_cib_version,
    get_cib,
    get_cib_xml,
    get_cluster_status_dom,
    push_cib_diff_xml,
    replace_cib_configuration,
    wait_for_idle,
)
from pcs.lib.pacemaker.values import get_valid_timeout_seconds
from pcs.lib.services import get_service_manager
from pcs.lib.tools import create_tmp_cib
from pcs.lib.xml_tools import etree_to_str

WaitType = Union[None, bool, int, str]


def _wait_type_to_int(wait: WaitType) -> int:
    """
    Convert WaitType to int.

    wait -- wait value. If False, it means wait is disabled, therefore -1
        is returned. If None, wait is enabled without timeout, therefore 0
        is returned. If string representing timeout or positive integer,
        wait is enabled with timeout, therefore number of seconds is
        retuned. Otherwise a LibraryError is raised.
    """
    if wait is False:
        return -1
    wait_timeout = get_valid_timeout_seconds(wait)
    if wait_timeout is None:
        return 0
    return wait_timeout


class LibraryEnvironment:
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods

    def __init__(
        self,
        logger: Logger,
        report_processor: reports.ReportProcessor,
        user_login: Optional[str] = None,
        user_groups: Optional[StringIterable] = None,
        cib_data: Optional[str] = None,
        corosync_conf_data: Optional[str] = None,
        booth_files_data: Optional[Mapping[str, Any]] = None,
        known_hosts_getter: Optional[
            Callable[[], Mapping[str, PcsKnownHost]]
        ] = None,
        # TODO there is no validation and transformation from str to int
        request_timeout: Optional[Union[int, str]] = None,
    ):
        # pylint: disable=too-many-arguments
        self._logger = logger
        self._report_processor = report_processor
        self._user_login = user_login
        self._user_groups = [] if user_groups is None else list(user_groups)
        self._cib_data = cib_data
        self._corosync_conf_data = corosync_conf_data
        self._booth_files_data = booth_files_data or {}
        self._request_timeout = request_timeout
        # TODO tokens probably should not be inserted from outside, but we're
        # postponing dealing with them, because it's not that easy to move
        # related code currently - it's in pcsd
        self._known_hosts_getter = known_hosts_getter
        self._known_hosts: Optional[Mapping[str, PcsKnownHost]] = None
        self._cib_upgrade_reported: bool = False
        self._cib_data_tmp_file: Optional[Any] = None  # TODO proper type hint
        self.__loaded_cib_diff_source: Optional[str] = None
        self.__loaded_cib_to_modify: Optional[_Element] = None
        self._communicator_factory = NodeCommunicatorFactory(
            LibCommunicatorLogger(self.logger, self.report_processor),
            self.user_login,
            self.user_groups,
            self._request_timeout,
        )
        self.__loaded_booth_env: Optional[BoothEnv] = None
        self.__loaded_dr_env: Optional[DrEnv] = None
        self.__service_manager: Optional[ServiceManagerInterface] = None

    @property
    def logger(self) -> Logger:
        return self._logger

    @property
    def report_processor(self) -> ReportProcessor:
        return self._report_processor

    @property
    def user_login(self) -> Optional[str]:
        return self._user_login

    @property
    def user_groups(self) -> Optional[list[str]]:
        return self._user_groups

    @property
    def ghost_file_codes(self) -> list[file_type_codes.FileTypeCode]:
        codes = set()
        if not self.is_cib_live:
            codes.add(file_type_codes.CIB)
        if not self.is_corosync_conf_live:
            codes.add(file_type_codes.COROSYNC_CONF)
        return sorted(codes)

    def get_cib(
        self,
        minimal_version: Optional[Version] = None,
        nice_to_have_version: Optional[Version] = None,
    ) -> _Element:
        if self.__loaded_cib_diff_source is not None:
            raise AssertionError("CIB has already been loaded")

        self.__loaded_cib_diff_source = get_cib_xml(self.cmd_runner())
        self.__loaded_cib_to_modify = get_cib(self.__loaded_cib_diff_source)

        if (
            nice_to_have_version is not None
            and minimal_version is not None
            and minimal_version >= nice_to_have_version
        ):
            nice_to_have_version = None

        for version, mandatory in (
            (nice_to_have_version, False),
            (minimal_version, True),
        ):
            if version is not None:
                upgraded_cib, was_upgraded = ensure_cib_version(
                    self.cmd_runner(),
                    self.__loaded_cib_to_modify,
                    version,
                    fail_if_version_not_met=mandatory,
                )
                if was_upgraded:
                    self.__loaded_cib_to_modify = upgraded_cib
                    self.__loaded_cib_diff_source = etree_to_str(upgraded_cib)
                    if not self._cib_upgrade_reported:
                        self.report_processor.report(
                            ReportItem.info(
                                reports.messages.CibUpgradeSuccessful()
                            )
                        )
                    self._cib_upgrade_reported = True

        return self.__loaded_cib_to_modify

    @property
    def cib(self) -> _Element:
        if self.__loaded_cib_to_modify is None:
            raise AssertionError("CIB has not been loaded")
        return self.__loaded_cib_to_modify

    def get_cluster_state(self) -> _Element:
        return get_cluster_status_dom(self.cmd_runner())

    def wait_for_idle(self, timeout: int = 0) -> None:
        """
        Wait for the cluster to settle down.

        timeout -- timeout in seconds, if less than 0 wait will be skipped, if 0
            wait indefinitely
        """
        if timeout < 0:
            # timeout is turned off
            return
        self.report_processor.report(
            ReportItem.info(reports.messages.WaitForIdleStarted(timeout))
        )
        wait_for_idle(self.cmd_runner(), timeout)

    def ensure_wait_satisfiable(self, wait: WaitType) -> int:
        """
        Convert WaitType to int. Returns wait timeout in seconds.
        Raise when wait is not supported or when wait is not valid wait value.

        wait -- wait value. If False, it means wait is disabled, therefore -1
            is returned. If None, wait is enabled without timeout, therefore 0
            is returned. If string representing timeout or positive integer,
            wait is enabled with timeout, therefore number of seconds is
            returned. Otherwise a LibraryError is raised.
        """
        timeout = _wait_type_to_int(wait)
        self._ensure_wait_satisfiable(timeout)
        return timeout

    def _ensure_wait_satisfiable(self, wait_timeout: int) -> None:
        if wait_timeout >= 0 and not self.is_cib_live:
            raise LibraryError(
                ReportItem.error(reports.messages.WaitForIdleNotLiveCluster())
            )

    def push_cib(self, custom_cib=None, wait_timeout: int = -1) -> None:
        """
        Push previously loaded instance of CIB or a custom CIB

        etree custom_cib -- push a custom CIB instead of a loaded instance
            (allows to push an externally provided CIB and replace the one in
            the cluster completely)
        wait_timeout -- wait timeout in seconds, if less than 0 wait will be
            skipped, if 0 wait indefinitely
        """
        self._ensure_wait_satisfiable(wait_timeout)
        if custom_cib is not None:
            if self.__loaded_cib_diff_source is not None:
                raise AssertionError(
                    "CIB has been loaded, cannot push custom CIB"
                )
            return self.__push_cib_full(custom_cib, wait_timeout)
        if self.__loaded_cib_diff_source is None:
            raise AssertionError("CIB has not been loaded")
        return self.__push_cib_diff(wait_timeout)

    def __push_cib_full(self, cib_to_push, wait_timeout: int):
        self.__do_push_cib(
            lambda: replace_cib_configuration(self.cmd_runner(), cib_to_push),
            wait_timeout,
        )

    def __push_cib_diff(self, wait_timeout: int):
        self.__do_push_cib(
            lambda: self.__main_push_cib_diff(self.cmd_runner()), wait_timeout
        )

    def __main_push_cib_diff(self, cmd_runner):
        cib_diff_xml = diff_cibs_xml(
            cmd_runner,
            self.report_processor,
            self.__loaded_cib_diff_source,
            etree_to_str(self.__loaded_cib_to_modify),
        )
        if cib_diff_xml:
            push_cib_diff_xml(cmd_runner, cib_diff_xml)

    def __do_push_cib(self, push_strategy, wait_timeout: int) -> None:
        push_strategy()
        self._cib_upgrade_reported = False
        self.__loaded_cib_diff_source = None
        self.__loaded_cib_to_modify = None
        if self.is_cib_live:
            self.wait_for_idle(wait_timeout)

    @property
    def is_cib_live(self) -> bool:
        return self._cib_data is None

    @property
    def final_mocked_cib_content(self) -> str:
        if self._cib_data is None:
            raise AssertionError(
                "Final mocked cib content does not make sense in live env."
            )

        if self._cib_data_tmp_file:
            self._cib_data_tmp_file.seek(0)
            return self._cib_data_tmp_file.read()

        return self._cib_data

    def get_corosync_conf_data(self) -> str:
        if self._corosync_conf_data is None:
            return get_local_corosync_conf()
        return self._corosync_conf_data

    def get_corosync_conf(self) -> CorosyncConfigFacade:
        # TODO The architecture of working with corosync.conf needs to be
        # overhauled to match the new file framework. The code below is
        # complicated, because we read corosync.conf data at one place outside
        # of the file framework or get it from outside (from CLI) and then we
        # put them back into the framework.
        corosync_instance = FileInstance.for_corosync_conf()
        try:
            facade = cast(
                CorosyncConfigFacade,
                corosync_instance.raw_to_facade(
                    self.get_corosync_conf_data().encode("utf-8")
                ),
            )
        except ParserErrorException as e:
            if self.report_processor.report_list(
                corosync_instance.toolbox.parser.exception_to_report_list(
                    e,
                    corosync_instance.toolbox.file_type_code,
                    (
                        corosync_instance.raw_file.metadata.path
                        if self.is_corosync_conf_live
                        else None
                    ),
                    force_code=None,
                    is_forced_or_warning=False,
                )
            ).has_errors:
                raise LibraryError() from e
        return facade

    def push_corosync_conf(
        self,
        corosync_conf_facade: CorosyncConfigFacade,
        skip_offline_nodes: bool = False,
    ) -> None:
        bad_sections, bad_attr_names, bad_attr_values = verify_corosync_section(
            corosync_conf_facade.config
        )
        if bad_sections or bad_attr_names or bad_attr_values:
            raise LibraryError(
                ReportItem.error(
                    reports.messages.CorosyncConfigCannotSaveInvalidNamesValues(
                        bad_sections,
                        bad_attr_names,
                        bad_attr_values,
                    )
                )
            )
        corosync_conf_data = corosync_conf_facade.config.export()
        if self.is_corosync_conf_live:
            node_name_list, report_list = get_existing_nodes_names(
                corosync_conf_facade,
                # Pcs is unable to communicate with nodes missing names. It
                # cannot send new corosync.conf to them. That might break the
                # cluster. Hence we error out.
                error_on_missing_name=True,
            )
            if self.report_processor.report_list(report_list).has_errors:
                raise LibraryError()

            self._push_corosync_conf_live(
                self.get_node_target_factory().get_target_list(
                    node_name_list,
                    skip_non_existing=skip_offline_nodes,
                ),
                corosync_conf_data,
                corosync_conf_facade.need_stopped_cluster,
                corosync_conf_facade.need_qdevice_reload,
                skip_offline_nodes,
            )
        else:
            self._corosync_conf_data = corosync_conf_data

    def _push_corosync_conf_live(
        self,
        target_list,
        corosync_conf_data,
        need_stopped_cluster,
        need_qdevice_reload,
        skip_offline_nodes,
    ):
        # TODO
        # * check for online nodes and run all commands on them only
        # * if those commands fail, exit with an error
        # * add support for allow_skip_offline=False
        # * use simple report processor
        # Correct reloading is done in pcs.lib.cluster.remove_nodes for example.

        # Check if the cluster is stopped when needed
        if need_stopped_cluster:
            com_cmd = CheckCorosyncOffline(
                self.report_processor, skip_offline_nodes
            )
            com_cmd.set_targets(target_list)
            run_and_raise(self.get_node_communicator(), com_cmd)
        # Distribute corosync.conf
        com_cmd = DistributeCorosyncConf(
            self.report_processor, corosync_conf_data, skip_offline_nodes
        )
        com_cmd.set_targets(target_list)
        run_and_raise(self.get_node_communicator(), com_cmd)
        # Reload corosync
        if not need_stopped_cluster:
            # If cluster must be stopped then we cannot reload corosync because
            # the cluster is stopped. If it is not stopped, we do not even get
            # here.
            com_cmd = ReloadCorosyncConf(self.report_processor)
            com_cmd.set_targets(target_list)
            run_and_raise(self.get_node_communicator(), com_cmd)
        # Reload qdevice if needed
        if need_qdevice_reload:
            self.report_processor.report(
                ReportItem.info(reports.messages.QdeviceClientReloadStarted())
            )
            com_cmd = qdevice.Stop(self.report_processor, skip_offline_nodes)
            com_cmd.set_targets(target_list)
            run(self.get_node_communicator(), com_cmd)
            has_errors = com_cmd.has_errors
            com_cmd = qdevice.Start(self.report_processor, skip_offline_nodes)
            com_cmd.set_targets(target_list)
            run(self.get_node_communicator(), com_cmd)
            has_errors = has_errors or com_cmd.has_errors
            if has_errors:
                raise LibraryError()

    @property
    def is_corosync_conf_live(self) -> bool:
        return self._corosync_conf_data is None

    def cmd_runner(
        self, env: Optional[Mapping[str, str]] = None
    ) -> CommandRunner:
        runner_env = {
            # make sure to get output of external processes in English and ASCII
            "LC_ALL": "C",
        }

        if self.user_login:
            runner_env["CIB_user"] = self.user_login

        if self._cib_data is not None:
            # Dump CIB data to a temporary file and set it up in the runner.
            # This way every called pacemaker tool can access the CIB and we
            # don't need to take care of it every time the runner is called.
            if not self._cib_data_tmp_file:
                self._cib_data_tmp_file = create_tmp_cib(
                    self.report_processor, self._cib_data
                )
            runner_env["CIB_file"] = self._cib_data_tmp_file.name

        if env:
            runner_env.update(env)

        return CommandRunner(self.logger, self.report_processor, runner_env)

    @property
    def communicator_factory(self) -> NodeCommunicatorFactory:
        return self._communicator_factory

    def get_node_communicator(
        self,
        request_timeout: Optional[int] = None,
    ) -> Communicator:
        return self.communicator_factory.get_communicator(
            request_timeout=request_timeout
        )

    def get_node_target_factory(self) -> NodeTargetLibFactory:
        return NodeTargetLibFactory(
            self.__get_known_hosts(), self.report_processor
        )

    def get_known_hosts(
        self, host_name_list: StringIterable
    ) -> list[PcsKnownHost]:
        known_hosts = self.__get_known_hosts()
        return [
            known_hosts[host_name]
            for host_name in host_name_list
            if host_name in known_hosts
        ]

    def __get_known_hosts(self) -> Mapping[str, PcsKnownHost]:
        if self._known_hosts is None:
            if self._known_hosts_getter:
                self._known_hosts = self._known_hosts_getter()
            else:
                self._known_hosts = {}
        return self._known_hosts

    def get_booth_env(self, name: Optional[str]) -> BoothEnv:
        if self.__loaded_booth_env is None:
            self.__loaded_booth_env = BoothEnv(name, self._booth_files_data)
        return self.__loaded_booth_env

    def get_dr_env(self) -> DrEnv:
        if self.__loaded_dr_env is None:
            self.__loaded_dr_env = DrEnv()
        return self.__loaded_dr_env

    def _get_service_manager(self) -> ServiceManagerInterface:
        if self.__service_manager is None:
            self.__service_manager = get_service_manager(
                self.cmd_runner(), self.report_processor
            )
        return self.__service_manager

    @property
    def service_manager(self) -> ServiceManagerInterface:
        return self._get_service_manager()



from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.common import reports
from pcs.common.services.interfaces import ServiceManagerInterface
from pcs.lib import cluster_property as lib_cluster_property
from pcs.lib.cib.tools import IdProvider
from pcs.lib.resource_agent.facade import ResourceAgentFacade
from pcs.lib.resource_agent.types import (
    ResourceAgentMetadata,
    ResourceAgentParameter,
)
from pcs.lib.xml_tools import etree_to_str

from pcs_test.tools import fixture
from pcs_test.tools.assertions import (
    assert_report_item_list_equal,
    assert_xml_equal,
)

TEMPLATE_CRM_CONFIG = """
<cib validate-with="pacemaker-3.8">
    <configuration>
        <crm_config>
            {property_sets}
        </crm_config>
    </configuration>
</cib>
"""

FIXTURE_TWO_PROPERTY_SETS = TEMPLATE_CRM_CONFIG.format(
    property_sets="""
        <cluster_property_set id="first">
            <nvpair id="first-have-watchdog" name="have-watchdog"
                value="false"/>
        </cluster_property_set>
        <cluster_property_set id="second">
            <nvpair id="second-maintenance-mode" name="maintenance-mode"
                value="false"/>
        </cluster_property_set>
    """
)


FIXTURE_NO_SET = TEMPLATE_CRM_CONFIG.format(property_sets="")

FORBIDDEN_OPTIONS_LIST = [
    "cluster-infrastructure",
    "cluster-name",
    "dc-version",
    "have-watchdog",
    "last-lrm-refresh",
]

PARAMETER_DEFINITIONS = [
    ("bool_param", "bool", "false", None),
    ("integer_param", "integer", "9", None),
    ("percentage_param", "percentage", "80%", None),
    ("select_param", "select", "s1", ["s1", "s2", "s3"]),
    ("time_param", "time", "30s", None),
    ("stonith-watchdog-timeout", "time", "0", None),
    ("cluster-infrastructure", "string", "corosync", None),
    ("cluster-name", "string", "(null)", None),
    ("dc-version", "string", "none", None),
    ("have-watchdog", "boolean", "false", None),
]

ALLOWED_PROPERTIES = [
    "bool_param",
    "integer_param",
    "percentage_param",
    "select_param",
    "stonith-watchdog-timeout",
    "time_param",
]

FIXTURE_VALID_OPTIONS_DICT = {
    "bool_param": "true",
    "integer_param": "10",
    "percentage_param": "20%",
    "select_param": "s3",
    "stonith-watchdog-timeout": "0",
    "time_param": "5min",
}

FIXTURE_INVALID_OPTIONS_DICT = {
    "bool_param": "Falsch",
    "integer_param": "3.14",
    "percentage_param": "20",
    "select_param": "not-in-enum-values",
    "time_param": "10x",
    "unknown": "value",
    "have-watchdog": "100",
}

STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES = ["", "0", "0s"]


def _fixture_parameter(name, param_type, default, enum_values):
    return ResourceAgentParameter(
        name,
        shortdesc=None,
        longdesc=None,
        type=param_type,
        default=default,
        enum_values=enum_values,
        required=False,
        advanced=False,
        deprecated=False,
        deprecated_by=[],
        deprecated_desc=None,
        unique_group=None,
        reloadable=False,
    )


FIXTURE_PARAMETER_LIST = [
    _fixture_parameter(*parameter) for parameter in PARAMETER_DEFINITIONS
]


FIXTURE_ERROR_REPORTS = [
    fixture.error(
        reports.codes.INVALID_OPTIONS,
        force_code=reports.codes.FORCE,
        option_names=["unknown"],
        allowed=ALLOWED_PROPERTIES,
        option_type="cluster property",
        allowed_patterns=[],
    ),
    fixture.error(
        reports.codes.INVALID_OPTIONS,
        force_code=None,
        option_names=["have-watchdog"],
        allowed=ALLOWED_PROPERTIES,
        option_type="cluster property",
        allowed_patterns=[],
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="integer_param",
        option_value="3.14",
        allowed_values="an integer or INFINITY or -INFINITY",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="percentage_param",
        option_value="20",
        allowed_values=(
            "a non-negative integer followed by '%' (e.g. 0%, 50%, "
            "200%, ...)"
        ),
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="select_param",
        option_value="not-in-enum-values",
        allowed_values=["s1", "s2", "s3"],
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
    fixture.error(
        reports.codes.INVALID_OPTION_VALUE,
        force_code=reports.codes.FORCE,
        option_name="time_param",
        option_value="10x",
        allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
        cannot_be_empty=False,
        forbidden_characters=None,
    ),
]


def warning_reports(report_list):
    warning_report_list = []
    for item_tuple in report_list:
        item_list = list(item_tuple)
        if item_list[3] is not None:
            item_list[0] = reports.ReportItemSeverity.WARNING
            item_list[3] = None
        warning_report_list.append(tuple(item_list))
    return warning_report_list


class TestValidateSetClusterProperties(TestCase):
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    def setUp(self):
        self.mock_service_manager = mock.Mock(spec=ServiceManagerInterface)
        self.mock_facade_list = [
            mock.Mock(
                spec=ResourceAgentFacade,
                metadata=mock.Mock(
                    spec=ResourceAgentMetadata,
                    parameters=FIXTURE_PARAMETER_LIST[
                        0 : int(len(FIXTURE_PARAMETER_LIST) / 2)
                    ],
                ),
            ),
            mock.Mock(
                spec=ResourceAgentFacade,
                metadata=mock.Mock(
                    spec=ResourceAgentMetadata,
                    parameters=FIXTURE_PARAMETER_LIST[
                        int(len(FIXTURE_PARAMETER_LIST) / 2) :
                    ],
                ),
            ),
        ]
        self.patcher_is_sbd_enabled = mock.patch("pcs.lib.sbd.is_sbd_enabled")
        self.patcher_sbd_devices = mock.patch(
            "pcs.lib.sbd.get_local_sbd_device_list"
        )
        self.patcher_sbd_timeout = mock.patch(
            "pcs.lib.sbd._get_local_sbd_watchdog_timeout"
        )
        self.mock_is_sbd_enabled = self.patcher_is_sbd_enabled.start()
        self.mock_sbd_devices = self.patcher_sbd_devices.start()
        self.mock_sbd_timeout = self.patcher_sbd_timeout.start()

    def tearDown(self):
        self.patcher_is_sbd_enabled.stop()
        self.patcher_sbd_devices.stop()
        self.patcher_sbd_timeout.stop()

    def assert_validate_set(
        self,
        configured_properties,
        new_properties,
        expected_report_list,
        sbd_enabled=False,
        sbd_devices=False,
        force=False,
        valid_value=True,
    ):
        self.mock_is_sbd_enabled.return_value = sbd_enabled
        self.mock_sbd_devices.return_value = ["devices"] if sbd_devices else []
        self.mock_sbd_timeout.return_value = 10
        assert_report_item_list_equal(
            lib_cluster_property.validate_set_cluster_properties(
                self.mock_facade_list,
                "property-set-id",
                configured_properties,
                new_properties,
                self.mock_service_manager,
                force=force,
            ),
            expected_report_list,
        )
        if (
            "stonith-watchdog-timeout" in new_properties
            and (
                new_properties["stonith-watchdog-timeout"]
                or "stonith-watchdog-timeout" in configured_properties
            )
            and valid_value
        ):
            self.mock_is_sbd_enabled.assert_called_once_with(
                self.mock_service_manager
            )
            if sbd_enabled:
                self.mock_sbd_devices.assert_called_once_with()
                if sbd_devices:
                    self.mock_sbd_timeout.assert_not_called()
                else:
                    if (
                        new_properties["stonith-watchdog-timeout"]
                        in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES
                    ):
                        self.mock_sbd_timeout.assert_not_called()
                    else:
                        self.mock_sbd_timeout.assert_called_once_with()
                self.mock_sbd_devices.reset_mock()
            else:
                self.mock_sbd_devices.assert_not_called()
                self.mock_sbd_timeout.assert_not_called()
        else:
            self.mock_is_sbd_enabled.assert_not_called()
            self.mock_sbd_devices.assert_not_called()
            self.mock_sbd_timeout.assert_not_called()

        self.mock_is_sbd_enabled.reset_mock()
        self.mock_sbd_devices.reset_mock()
        self.mock_sbd_timeout.reset_mock()

    def test_no_properties_to_set_or_unset(self):
        self.assert_validate_set(
            [],
            {},
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                )
            ],
        )

    def test_no_properties_to_set_or_unset_forced(self):
        self.assert_validate_set(
            [],
            {},
            [
                fixture.warn(
                    reports.codes.ADD_REMOVE_ITEMS_NOT_SPECIFIED,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                )
            ],
            force=True,
        )

    def test_set_valid_properties_and_values(self):
        self.assert_validate_set([], FIXTURE_VALID_OPTIONS_DICT, [])

    def test_set_invalid_properties_and_values(self):
        self.assert_validate_set(
            [], FIXTURE_INVALID_OPTIONS_DICT, FIXTURE_ERROR_REPORTS
        )

    def test_set_invalid_properties_and_values_forced(self):
        self.assert_validate_set(
            [],
            FIXTURE_INVALID_OPTIONS_DICT,
            warning_reports(FIXTURE_ERROR_REPORTS),
            force=True,
        )

    def test_unset_stonith_watchdog_timeout_sbd_disabled(self):
        for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
            with self.subTest(value=value):
                self.assert_validate_set(
                    ["stonith-watchdog-timeout"],
                    {"stonith-watchdog-timeout": value},
                    [],
                )

    def test_set_stonith_watchdog_timeout_sbd_disabled(self):
        self.assert_validate_set(
            [],
            {"stonith-watchdog-timeout": "5"},
            [
                fixture.error(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
                    reason="sbd_not_set_up",
                )
            ],
        )

    def test_set_ok_stonith_watchdog_timeout_sbd_enabled_without_devices(self):
        for value in ["15", "15s"]:
            with self.subTest(value=value):
                self.assert_validate_set(
                    [],
                    {"stonith-watchdog-timeout": value},
                    [],
                    sbd_enabled=True,
                )

    def test_set_small_stonith_watchdog_timeout_sbd_enabled_without_devices(
        self,
    ):
        self.assert_validate_set(
            [],
            {"stonith-watchdog-timeout": "9s"},
            [
                fixture.error(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_TOO_SMALL,
                    force_code=reports.codes.FORCE,
                    cluster_sbd_watchdog_timeout=10,
                    entered_watchdog_timeout="9s",
                )
            ],
            sbd_enabled=True,
        )

    def test_set_small_stonith_watchdog_timeout_sbd_enabled_without_devices_forced(
        self,
    ):
        self.assert_validate_set(
            [],
            {"stonith-watchdog-timeout": "9"},
            [
                fixture.warn(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_TOO_SMALL,
                    cluster_sbd_watchdog_timeout=10,
                    entered_watchdog_timeout="9",
                )
            ],
            sbd_enabled=True,
            force=True,
        )

    def _set_invalid_value_stonith_watchdog_timeout(
        self, sbd_enabled=False, sbd_devices=False
    ):
        for value in ["invalid", "10x"]:
            with self.subTest(value=value):
                self.assert_validate_set(
                    [],
                    {"stonith-watchdog-timeout": value},
                    [
                        fixture.error(
                            reports.codes.INVALID_OPTION_VALUE,
                            option_name="stonith-watchdog-timeout",
                            option_value=value,
                            allowed_values="time interval (e.g. 1, 2s, 3m, 4h, ...)",
                            cannot_be_empty=False,
                            forbidden_characters=None,
                        )
                    ],
                    sbd_enabled=sbd_enabled,
                    sbd_devices=sbd_devices,
                    valid_value=False,
                )

    def test_set_invalid_value_stonith_watchdog_timeout_sbd_enabled_without_devices(
        self,
    ):
        self._set_invalid_value_stonith_watchdog_timeout(
            sbd_enabled=True, sbd_devices=False
        )

    def test_set_invalid_value_stonith_watchdog_timeout_sbd_enabled_with_devices(
        self,
    ):
        self._set_invalid_value_stonith_watchdog_timeout(
            sbd_enabled=True, sbd_devices=True
        )

    def test_set_invalid_value_stonith_watchdog_timeout_sbd_disabled(
        self,
    ):
        self._set_invalid_value_stonith_watchdog_timeout(
            sbd_enabled=False, sbd_devices=False
        )

    def test_unset_stonith_watchdog_timeout_sbd_enabled_without_devices(
        self,
    ):
        for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
            with self.subTest(value=value):
                self.assert_validate_set(
                    ["stonith-watchdog-timeout"],
                    {"stonith-watchdog-timeout": value},
                    [
                        fixture.error(
                            reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                            force_code=reports.codes.FORCE,
                            reason="sbd_set_up_without_devices",
                        )
                    ],
                    sbd_enabled=True,
                )

    def test_unset_stonith_watchdog_timeout_sbd_enabled_without_devices_forced(
        self,
    ):
        for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
            with self.subTest(value=value):
                self.assert_validate_set(
                    ["stonith-watchdog-timeout"],
                    {"stonith-watchdog-timeout": value},
                    [
                        fixture.warn(
                            reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_UNSET,
                            reason="sbd_set_up_without_devices",
                        )
                    ],
                    force=True,
                    sbd_enabled=True,
                )

    def test_set_stonith_watchdog_timeout_sbd_enabled_with_devices(self):
        self.assert_validate_set(
            [],
            {"stonith-watchdog-timeout": "15"},
            [
                fixture.error(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
                    force_code=reports.codes.FORCE,
                    reason="sbd_set_up_with_devices",
                )
            ],
            sbd_enabled=True,
            sbd_devices=True,
        )

    def test_set_stonith_watchdog_timeout_sbd_enabled_with_devices_forced(self):
        self.assert_validate_set(
            [],
            {"stonith-watchdog-timeout": "15s"},
            [
                fixture.warn(
                    reports.codes.STONITH_WATCHDOG_TIMEOUT_CANNOT_BE_SET,
                    reason="sbd_set_up_with_devices",
                )
            ],
            sbd_enabled=True,
            sbd_devices=True,
            force=True,
        )

    def test_unset_stonith_watchdog_timeout_sbd_enabled_with_devices(self):
        for value in STONITH_WATCHDOG_TIMEOUT_UNSET_VALUES:
            with self.subTest(value=value):
                self.assert_validate_set(
                    ["stonith-watchdog-timeout"],
                    {"stonith-watchdog-timeout": value},
                    [],
                    sbd_enabled=True,
                    sbd_devices=True,
                )

    def test_remove_not_configured_stonith_watchdog_timeout(self):
        self.assert_validate_set(
            ["a", "b"],
            {"stonith-watchdog-timeout": ""},
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["stonith-watchdog-timeout"],
                )
            ],
        )

    def test_remove_valid_configured_options(self):
        self.assert_validate_set(
            ["bool_param", "integer_param", "percentage_param", "a", "b", "c"],
            {"bool_param": "", "integer_param": ""},
            [],
        )

    def test_remove_valid_not_configured_options(self):
        self.assert_validate_set(
            [],
            {"bool_param": "", "integer_param": ""},
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["bool_param", "integer_param"],
                )
            ],
        )

    def test_remove_valid_not_configured_options_forced(self):
        self.assert_validate_set(
            [],
            {"bool_param": "", "integer_param": ""},
            [
                fixture.warn(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["bool_param", "integer_param"],
                )
            ],
            force=True,
        )

    def test_remove_invalid_configured_options(self):
        self.assert_validate_set(
            ["bool_param", "integer_param", "percentage_param", "a", "b", "c"],
            {"a": "", "b": ""},
            [],
        )

    def test_remove_invalid_configured_options_forced(self):
        self.assert_validate_set(
            ["bool_param", "integer_param", "percentage_param", "a", "b", "c"],
            {"a": "", "b": ""},
            [],
            force=True,
        )

    def test_remove_invalid_not_configured_options(self):
        self.assert_validate_set(
            ["a", "b", "c"],
            {"x": "", "y": ""},
            [
                fixture.error(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    force_code=reports.codes.FORCE,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["x", "y"],
                ),
            ],
        )

    def test_remove_invalid_not_configured_options_forced(self):
        self.assert_validate_set(
            ["a", "b", "c"],
            {"x": "", "y": ""},
            [
                fixture.warn(
                    reports.codes.ADD_REMOVE_CANNOT_REMOVE_ITEMS_NOT_IN_THE_CONTAINER,
                    container_type=reports.const.ADD_REMOVE_CONTAINER_TYPE_PROPERTY_SET,
                    item_type=reports.const.ADD_REMOVE_ITEM_TYPE_PROPERTY,
                    container_id="property-set-id",
                    item_list=["x", "y"],
                ),
            ],
            force=True,
        )

    def test_remove_forbidden_options(self):
        self.assert_validate_set(
            FORBIDDEN_OPTIONS_LIST,
            {key: "" for key in FORBIDDEN_OPTIONS_LIST},
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=FORBIDDEN_OPTIONS_LIST,
                    allowed=[
                        "bool_param",
                        "integer_param",
                        "percentage_param",
                        "select_param",
                        "stonith-watchdog-timeout",
                        "time_param",
                    ],
                    option_type="cluster property",
                    allowed_patterns=[],
                )
            ],
        )

    def test_remove_forbidden_options_forced(self):
        self.assert_validate_set(
            FORBIDDEN_OPTIONS_LIST,
            {key: "" for key in FORBIDDEN_OPTIONS_LIST},
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    option_names=FORBIDDEN_OPTIONS_LIST,
                    allowed=[
                        "bool_param",
                        "integer_param",
                        "percentage_param",
                        "select_param",
                        "stonith-watchdog-timeout",
                        "time_param",
                    ],
                    option_type="cluster property",
                    allowed_patterns=[],
                )
            ],
            force=True,
        )


class TestGetClusterPropertySetElementLegacy(TestCase):
    def test_return_first_set(self):
        # pylint: disable=no-self-use
        cib = etree.fromstring(FIXTURE_TWO_PROPERTY_SETS)
        id_provider = IdProvider(cib)
        set_element = (
            lib_cluster_property.get_cluster_property_set_element_legacy(
                cib, id_provider
            )
        )
        assert_xml_equal(
            etree_to_str(cib.xpath('.//*[@id="first"]')[0]),
            etree_to_str(set_element),
        )

    def test_no_set_create_cib_bootstrap_options_set(self):
        # pylint: disable=no-self-use
        cib = etree.fromstring(FIXTURE_NO_SET)
        id_provider = IdProvider(cib)
        set_element = (
            lib_cluster_property.get_cluster_property_set_element_legacy(
                cib, id_provider
            )
        )
        assert_xml_equal(
            '<cluster_property_set id="cib-bootstrap-options"/>',
            etree_to_str(set_element),
        )
